/* (C) Copyright 1984,85,86,87 Walter L. Peacock   All Rights Reserved    */
/*  g i v e i d x r . c  8/02/84  */

/*---------------------------------------------------------------------------- 
 * giveidxr() is used to allocate index pages and/or data records.  It 
 * uses the previously initialized header record to allocate new records.   It 
 * will allocate from the deleted record chain first if there are deleted 
 * records in the file otherwise records are allocated from the end of the 
 * file.  The header record consists of eight bytes as follows:  bytes 0-3 is 
 * the deleted record chain; bytes 4-7 is the end-of-file pointer.
 * 
 * When the filename is present in the btcommo member dtaname, then this 
 * function will open the file, allocate the new record, then close the 
 * file.  This technique should be used if you seem to be experiencing 
 * buffering problems or in multiuser applications when the header record 
 * update must be forced to the disk.
 * 
 * If your application will issue INSERT and/or DELETE calls, then
 * all record allocations automatically occur through calls to giveidxr() to
 * ensure proper maintenance of the file's header record.
 *---------------------------------------------------------------------------*/

#include "cbtree.h"

long giveidxr(fd, reclen, locrec, lochdr)
int locrec, lochdr, fd;
unsigned reclen;
{
   extern int open(), close();
   extern char *calloc();
   extern void free();
   BTBLKHDR freespcb;
   long lbuff = 0L;      /* prevent Lattice warning */
   long int pos, ret_loc, lseek();
   char chrbuf[sizeof(long)+2];
   int lk;

   ret_loc = 0L;

   if (lochdr == 1){	/* header record was not already locked so lock here. */
      lk = 0;
      while (lockunlk(fd, 0L, sizeof(BTBLKHDR), LOCKR) != 1)
         if (++lk > HDRTRYS)
            goto giv_0exit;        /* i.e. unable to lock header record */
	}

   getfhdr(&freespcb, fd);
   /* the following may also need to check for valid eoflst */
   if( (freespcb.freelst < 2) || (freespcb.freelst >= freespcb.eoflst) ) {
      /* allocate from eof */
      ret_loc = freespcb.eoflst++;
      freespcb.freelst = 0L;
      putfhdr(fd, &freespcb);

/* AmigaDOS does not allow you to seek beyond the end of a file.         */
/* This is normally NOT a problem.   However, if you desire to have the   */
/* index tree and the data records in the same file (see BTSETUP         */
/* documentation for further restrictions on this) it becomes necessary */
/* to pre-extend the file, but only if the free list is empty.         */
#if AmigaDOS
#if PRE_EXTEND
      if ((cp = calloc(reclen, SZCHAR)) == NULL)
         ckerror(- CK_NOMEM, "105");
      if (btseek(fd, ret_loc, reclen) == -1)
         ckerror(- CK_SEEK, "106");
      if (write(fd, cp, reclen) == ERR)
         ckerror(- CK_WRIT, "107");
      FREE(cp);
#endif
#endif
   }
   else {                        /* alloc and update freespcb    */
      ret_loc = freespcb.freelst;
      /* position pointer for previously deallocated records   */
      pos = (freespcb.freelst - 1L) * reclen;
      lk = 0;
      while (lockunlk(fd, pos, reclen, LOCKR) != 1)
         if (++lk > HDRTRYS){
            ret_loc = 0L;
            goto giv_1exit;  /* i.e. unable to lock free link record */
         }
      if (lseek(fd, pos, 0) == -1L)
         ckerror(- CK_SEEK, "32");
      if (read(fd, chrbuf, sizeof(long)+2) == ERR )
         ckerror(- CK_READ, "31");
      /* verify that record to allocate is a deleted record */
      if ( (chrbuf[0] == '~') && (chrbuf[5] == '~') ){
         lseek(fd, -5L, 1);   /* pos back to freelst chain link */
         if (read(fd, (char *)&lbuff, sizeof(long) ) == ERR)
            ckerror(- CK_READ, "33");
         freespcb.freelst = lbuff; /* maintain freelst chain */
         if( (freespcb.freelst < 2) ||
               (freespcb.freelst >= freespcb.eoflst) )
            freespcb.freelst = 0L;
      }
      else{ /* record is not marked as being a deleted record */
            /* so don't chance; allocate from eof */
         ret_loc = freespcb.eoflst++;
         freespcb.freelst = 0L;
      }
      putfhdr(fd, &freespcb);
      lockunlk(fd, (long) pos, reclen, UNLOCKR);
   }
   if(locrec == 1){
      lk = 0;
      while(lockunlk(fd, (ret_loc - 1L) * reclen, reclen, LOCKR) != 1)
         if (++lk > HDRTRYS)
            ret_loc = 0;   /* i.e. unable to lock allocated record */
   }
giv_1exit:
   if (lochdr == 1)	  /* header record was locked in this function so unlock. */
      lockunlk(fd, 0L, sizeof(BTBLKHDR), UNLOCKR); /* now unlock header rec. */
giv_0exit:
   return(ret_loc);
}
